import java.io.*;

/** This class open the NES file and start the emulator.
    This class needs a cleaning.
    This code is GPL.
    Copyright (c) 2000 Brice Fines.<DL>
    <UR>
    <LI>This program is free software; you can redistribute it and/or modify<BR>
    it under the terms of the GNU General Public License as published by<BR>
    the Free Software Foundation; either version 2 of the License, or<BR>
    (at your option) any later version.<BR>
    <LI>This program is distributed in the hope that it will be useful,<BR>
    but WITHOUT ANY WARRANTY; without even the implied warranty of<BR>
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the<BR>
    GNU General Public License for more details.<BR>
    <LI>You should have received a copy of the GNU General Public License<BR>
    along with this program; if not, write to the Free Software<BR>
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA<BR>
    </UR></DL>
    @version 0.2005 06/06/00
    @author <A HREF="mailto:bfines@uni2.es">Jarod CANAL</A>
    @see "The Gnu Public License." */
public final class Javanese{

    /** Main function. */
    public static void main( String[] args ) throws Exception {

	/** File input stream used to read file. */
	FileInputStream Fis;

	/** Array used for reading file input stream. */
	byte B[] = new byte[8];

	/** Number of ROM pages. */
	int PrgRomCount;

	/** Number of VROM pages. */
	int ChrRomCount;

	/** NES cart mapper. */
	int Mapper;

	/** Indicates is cart has a vertical or horizontal mirroring. <BR>
	    True = horizontal mirroring.
	    False = vertical mirroring. */
	boolean HasHorizontalMirroring;

	/** Indicates if cart has save RAM. <BR>
	    True : cart has save RAM.
	    False : cart does not have save RAM.*/
	boolean HasSRAM = false;

	/** NES. */
	NintendoEntertainmentSystem Nes /*= new NintendoEntertainmentSystem()*/;

	try { 
	    Fis = new FileInputStream( args[0] );
	    Fis.read( B );
	    if(  ( B[ 0 ] != 78 )||( B[ 1 ] != 69 )||( B[ 2 ] != 83 )||( B[ 3 ] != 26 ) ) System.out.println( "Bad file header: NES^Z" );
	    PrgRomCount = B[4];
	    ChrRomCount = B[5];
	    Mapper = ( B[6] >> 4 ) | ( B[7] & 0xf0 );
	    if ( ( B[7] & 0x0f ) != 0 ) 
		{
		    System.out.println( "Bad file header: non zero value in four lower bits of byte 7");
		    System.out.println( "These bits are reserved for future use, please repair your header.");
		    java.lang.System.exit(0);
		}
	    System.out.println( "Mapper: " + Mapper );
	    System.out.println( PrgRomCount * 16 + " K/ "+ ChrRomCount * 8 + " K" + "	( " + PrgRomCount + " ROM page(s), " + ChrRomCount + " VROM page(s) )" );
	    if ( ( B[ 6 ] & 0x01 ) == 0 ){
		System.out.println( "Horizontal Mirroring" );
		HasHorizontalMirroring = true;
	    }
	    else{
		System.out.println( "Vertical Mirroring" );
		HasHorizontalMirroring = false;
	    }
	    if ( ( B[ 6 ] & 0x02 ) != 0 ){
		System.out.println("SRAM enabled at $6000-$7FFF");
		HasSRAM = true;
	    }
	    if ( ( B[ 6 ] & 0x04 ) != 0 ) System.out.println( "512-byte trainer present at $7000-$71FF" );
	    if ( ( B[ 6 ] & 0x08 ) != 0 ) System.out.println( "Four-screen VRAM layout" );
	    Fis.read( B );
	    for ( int i = 0; i < 8; i ++ ) 
		{
		    if ( B[ i ] != 0 ) {
			System.out.println( "Bad file header: non zero value in byte " + i );
			System.out.println( "Bytes 8 to 15 are reserved for future use, please repair your header." );
			java.lang.System.exit( 0 );
		    }
		}
	    Nes = new NintendoEntertainmentSystem( PrgRomCount, ChrRomCount, HasHorizontalMirroring );
	    for ( int i = 0; i < ( 0x800 * PrgRomCount ) ; i ++ ) // $4000 / 8 = $800
		{
		    Fis.read( B );
		    for ( int j = 0; j < 8; j ++ ) Nes.RomImage[ i * 8 + j ] = ( char ) ( B[ j ] & 0x00FF );
		}
	    for ( int i = 0; i < ( 0x400 * ChrRomCount ) ; i ++ ) // $2000 / 8 = $400
		{
		    Fis.read( B );
		    for ( int j = 0; j < 8; j ++ ) Nes.VRomImage[ i * 8 + j ] = ( char ) ( B[ j ] & 0x00FF );
		}
	    Fis.close();

	    switch( Mapper )
		{
		case 0 : // Mapper 0, none
		    System.out.println("Mapper 0 : no mapper	( partially supported )");
		    if ( PrgRomCount != 1 && PrgRomCount != 2 )
			{
			    System.out.println( "Invalid number of 16k ROM page, not good." );
			    System.exit( 0 );
			}
		    if ( ChrRomCount == 0 )
			{
			    System.out.println( "Cart has no VROM, not good." );
			    System.exit( 0 );
			}
		    if ( ChrRomCount > 1 )
			{
			    System.out.println( "Cart has more than 8k VROM, not good." );
			    System.exit( 0 );
			}
		    if ( PrgRomCount == 1 )
			{
			    System.arraycopy( Nes.RomImage, 0, Nes.Memory, 0x8000, 0x4000 );
			    System.arraycopy( Nes.RomImage, 0, Nes.Memory, 0xC000, 0x4000 );
			}
		    else
			{
			    System.arraycopy( Nes.RomImage, 0, Nes.Memory, 0x8000, 0x8000 );
			}
		    System.arraycopy( Nes.VRomImage, 0, Nes.PPU.Memory, 0, 0x2000 );
		    Nes.PPU.caching();
		    break;
		    /*case 1 : // Mapper 1
		      Nes = new NintendoEntertainmentSystemWithMapper1();
		      break;*/
		default : // Unsupported mapper
		    System.out.println("Unsupported mapper :" + Mapper);
		    java.lang.System.exit( 0 );
		}


	    Nes.reset();
	    Nes.nes_timeslice();
	    System.out.println( "Thank you for using Javanese." );
	    java.lang.System.exit( 0 );

	}

	catch( IOException ioe )
	    {
		System.out.println( "NES file Read failure" );
		java.lang.System.exit( 0 );
	    }
	catch( ArrayIndexOutOfBoundsException aioobe )
	    {
		// This occurs when file.nes is not specified in command line.
		System.out.println( "Usage: java Javanese File.nes " );
		java.lang.System.exit( 0 );
	    }
    }
}
